from dataclasses import dataclass

from typing import List, Optional

from gbso.cpu.cpu import CPU
from gbso.cpu.state import CPUState
from gbso.cpu.insn import Insn


@dataclass
class Program:
    insns: List[Insn]

    def execute(self, init_state: Optional[CPUState] = None) -> CPU:
        cpu = CPU(state=init_state)
        for insn in self.insns:
            insn.exec(cpu)
        return cpu

    def perf(self) -> int:
        return sum([insn.cycles() for insn in self.insns])
